# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccount_OAuth.ipynb.

# %% auto 0
__all__ = ['DomoAccountOAuth_Config_SnowflakeOauth', 'DomoAccountOAuth_Config_JiraOnPremOauth', 'OAuthConfig',
           'DomoAccount_OAuth']

# %% ../../nbs/classes/50_DomoAccount_OAuth.ipynb 3
from dataclasses import dataclass, field
import httpx
from typing import Any

import domolibrary.client.DomoAuth as dmda

import domolibrary.routes.account as account_routes
import domolibrary.classes.DomoAccount_Default as dmacb
import domolibrary.classes.DomoAccount_Config as dmacnfg
import domolibrary.classes.DomoAccess as dmacc

from nbdev.showdoc import patch_to

# %% ../../nbs/classes/50_DomoAccount_OAuth.ipynb 7
@dataclass
class DomoAccountOAuth_Config_SnowflakeOauth(dmacnfg.DomoAccount_Config):
    data_provider_type: str = "snowflake-oauth-config"
    is_oauth: bool = True

    client_id: str = None
    secret: str = None

    @classmethod
    def _from_json(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            client_id=obj["client-id"],
            secret=obj["client-secret"],
            raw=obj,
            parent=parent,
        )

    def to_json(self):
        return {"client-id": self.client_id, "client-secret": self.secret}


@dataclass
class DomoAccountOAuth_Config_JiraOnPremOauth(dmacnfg.DomoAccount_Config):
    data_provider_type: str = "jira-on-prem-oauth-config"
    is_oauth: bool = True

    client_id: str = None
    secret: str = None

    @classmethod
    def _from_json(cls, obj: dict, parent: Any = None):
        return cls(
            client_id=obj["client_id"],
            secret=obj["client_secret"],
            raw=obj,
            parent=parent,
        )

    def to_json(self):
        return {"client_id": self.client_id, "client_secret": self.secret}


class OAuthConfig(dmacnfg.ConfigEnum):
    snowflake_oauth_config = DomoAccountOAuth_Config_SnowflakeOauth

    jira_on_prem_oauth_config = DomoAccountOAuth_Config_JiraOnPremOauth

    @classmethod
    def _missing_(cls, value):
        alt_search_str = cls.generate_alt_search_str(value)

        config_match = next(
            (member for member in cls if member.name in [value, alt_search_str]),
            None,
        )

        ## best case scenario alt_search yields a result
        if not config_match:
            raise dmacnfg.AccountConfig_ProviderTypeNotDefined(value)

        return config_match

# %% ../../nbs/classes/50_DomoAccount_OAuth.ipynb 9
@dataclass
class DomoAccount_OAuth(dmacb.DomoAccount_Default):
    Access: dmacc.DomoAccess_OAuth = field(repr=False, default=None)

    def __post_init__(self):
        self.Access = dmacc.DomoAccess_OAuth(auth=self.auth, parent=self)

    async def _get_config(
        self,
        session=None,
        return_raw: bool = False,
        debug_api: bool = None,
        debug_num_stacks_to_drop=2,
        is_suppress_no_config: bool = False,  # can be used to suppress cases where the config is not defined, either because the account_config is OAuth, and therefore not stored in Domo OR because the AccountConfig class doesn't cover the data_type
    ):
        if not self.data_provider_type:
            res = await account_routes.get_account_by_id(
                auth=self.auth,
                account_id=self.id,
                session=session,
                debug_api=debug_api,
                parent_class=self.__class__.__name__,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            )

            self.data_provider_type = res.response["dataProviderType"]

        res = await account_routes.get_oauth_account_config(
            auth=self.auth,
            account_id=self.id,
            session=session,
            debug_api=debug_api,
            data_provider_type=self.data_provider_type,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        config_fn = OAuthConfig(self.data_provider_type).value

        if not is_suppress_no_config and not config_fn.is_defined_config:
            raise config_fn._associated_exception(self.data_provider_type)

        self.Config = config_fn._from_json(res.response)

        if self.Config and self.Config.to_json() != {}:
            if not res.response:
                print(self.data_provider_type, "no response")

            if not self.Config.to_json():
                print(
                    self.id,
                    self.data_provider_type,
                    "no config",
                    self.Config.to_json(),
                    res.response,
                )

            self._test_missing_keys(
                res_obj=res.response, config_obj=self.Config.to_json()
            )

        return self.Config

    @classmethod
    async def get_by_id(
        cls,
        auth: dmda.DomoAuth,
        account_id: int,
        is_suppress_no_config: bool = True,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        **kwargs
    ):
        """retrieves account metadata and attempts to retrieve config"""

        res = await account_routes.get_oauth_account_by_id(
            auth=auth,
            account_id=account_id,
            session=session,
            debug_api=debug_api,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        acc = cls._from_json(
            obj=res.response,
            auth=auth,
            is_admin_summary=False,
            is_use_default_account_class=False,
            **kwargs
        )

        await acc._get_config(
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
            is_suppress_no_config=is_suppress_no_config,
        )

        return acc

    @classmethod
    async def create(
        cls,
        auth: dmda.DomoAuth,
        account_name: str,
        oauth_config: OAuthConfig,
        origin: str = "OAUTH_CONFIGURATION",
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
    ):
        res = await account_routes.create_oauth_account(
            account_name=account_name,
            data_provider_type=oauth_config.data_provider_type,
            origin=origin,
            config=oauth_config.to_json(),
            debug_api=debug_api,
            session=session,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        return await cls.get_by_id(
            auth=auth,
            account_id=res.response["id"],
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

    async def delete(
        self,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        return await account_routes.delete_oauth_account(
            auth=self.auth,
            account_id=self.id,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

    async def update_name(
        self,
        account_name: str,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient = None,
    ):
        await account_routes.update_oauth_account_name(
            auth=self.auth,
            account_id=self.id,
            account_name=account_name,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

        self.name = account_name
        return self

    async def update_config(
        self,
        oauth_config: OAuthConfig = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        await account_routes.update_oauth_account_config(
            auth=self.auth,
            account_id=self.id,
            config_body=oauth_config.to_json() or self.Config.to_json(),
            data_provider_type=self.data_provider_type,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

        self.Config = oauth_config

        return self



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoBootstrap.ipynb.

# %% auto 0
__all__ = ['DomoBootstrap_Feature', 'DomoBootstrap']

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 2
from ..routes.bootstrap import InvalidAuthTypeError

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 3
from dataclasses import dataclass, field
from nbdev.showdoc import patch_to
import httpx

import domolibrary.classes.DomoPage as dmpg

import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.routes.bootstrap as bootstrap_routes
import domolibrary.utils.chunk_execution as ce

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 5
@dataclass
class DomoBootstrap_Feature:
    id: int
    name: str
    label: str
    type: str
    purchased: bool
    enabled: bool

    @classmethod
    def _from_json_bootstrap(cls, json_obj: dict):
        dd = util_dd.DictDot(json_obj)

        bsf = cls(
            id=dd.id,
            name=dd.name,
            label=dd.label,
            type=dd.type,
            purchased=dd.purchased,
            enabled=dd.enabled,
        )
        return bsf

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 6
@dataclass
class DomoBootstrap:
    auth: dmda.DomoAuth = field(repr=False)
    bootstrap: dict = field(default=None)
    customer_id: str = None
    page_ls: list[dmpg.DomoPage] = field(default=None)
    feature_ls: list[DomoBootstrap_Feature] = field(default=None)

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 7
@patch_to(DomoBootstrap)
async def get_all(
    self: DomoBootstrap,
    auth: dmda.DomoFullAuth = None,
    debug_api: bool = False,
    return_raw: bool = False,
    debug_num_stacks_to_drop=2,
):
    auth = auth or self.auth

    res = await bootstrap_routes.get_bootstrap(
        auth=auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__.__name__,
    )

    self.bootstrap = res.response

    if return_raw:
        return res

    return res.response

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 10
@patch_to(DomoBootstrap)
async def get_customer_id(
    self: DomoBootstrap,
    auth: dmda.DomoFullAuth = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=3,
    return_raw: bool = False,
):
    res = await bootstrap_routes.get_bootstrap_customerid(
        auth=auth or self.auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        return_raw=return_raw,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res

    self.customer_id = res.response

    return self.customer_id

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 13
@patch_to(DomoBootstrap)
async def get_pages(
    self: DomoBootstrap,
    auth: dmda.DomoFullAuth = None,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
) -> list[dmpg.DomoPage]:
    auth = auth or self.auth

    res = await bootstrap_routes.get_bootstrap_pages(
        auth=auth,
        debug_api=debug_api,
        parent_class=self.__class__.__name__,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        return None

    self.page_ls = await ce.gather_with_concurrency(
        n=60,
        *[
            dmpg.DomoPage._from_bootstrap(page_obj, auth=auth)
            for page_obj in res.response
        ]
    )

    return self.page_ls

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 16
@patch_to(DomoBootstrap)
async def get_features(
    self: DomoBootstrap,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    return_raw: bool = False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
):
    auth = auth or self.auth

    res = await bootstrap_routes.get_bootstrap_features(
        auth=auth,
        session=session,
        debug_api=debug_api,
        return_raw=return_raw,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res

    feature_list = [
        DomoBootstrap_Feature._from_json_bootstrap(json_obj)
        for json_obj in res.response
    ]

    return feature_list

# %% ../../nbs/classes/50_DomoBootstrap.ipynb 19
@patch_to(DomoBootstrap)
async def is_feature_accountsv2_enabled(
    self: DomoBootstrap,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    return_raw: bool = False,
    debug_num_stacks_to_drop=3,
):
    res = await bootstrap_routes.get_bootstrap_features_is_accountsv2_enabled(
        auth=auth or self.auth,
        return_raw=return_raw,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res

    return res.response



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccount.ipynb.

# %% auto 0
__all__ = ['DomoAccount', 'DomoAccount_Types', 'DomoAccounts_NoAccount', 'DomoAccounts']

# %% ../../nbs/classes/50_DomoAccount.ipynb 3
from domolibrary.routes.account import (
    Account_GET_Error,
    Account_NoMatch,
    Account_CRUD_Error,
    Account_CreateParams_Error,
    Account_Config_Error,
    ShareAccount,
    ShareAccount_V1_AccessLevel,
    ShareAccount_AccessLevel,
    Account_Share_Error,
    Account_AlreadyShared_Error,
)

from domolibrary.classes.DomoAccount_Config import (
    AccountConfig_UsesOauth,
    AccountConfig_ProviderTypeNotDefined,
    DomoAccount_Config,
    AccountConfig,
)

from domolibrary.classes.DomoAccount_Default import (
    Account_CanIModify,
    UpsertAccount_MatchCriteria,
    DomoAccount_Default,
    DomoAccounConfig_MissingFields,
)

from domolibrary.classes.DomoAccount_Credential import (
    DAC_NoTargetInstance,
    DAC_NoPassword,
    DAC_NoUserName,
    DAC_NoAccessToken,
    DAC_ValidAuth,
    DomoAccount_Credential,
)

from domolibrary.classes.DomoAccount_OAuth import (
    DomoAccount_OAuth,
    OAuthConfig
)

# %% ../../nbs/classes/50_DomoAccount.ipynb 4
from dataclasses import dataclass, field
from typing import List
from enum import Enum

import httpx
import datetime as dt
from copy import deepcopy

from nbdev.showdoc import patch_to

import domolibrary.routes.account as account_routes
import domolibrary.routes.datacenter as datacenter_routes

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde
import domolibrary.utils.chunk_execution as dmce

# %% ../../nbs/classes/50_DomoAccount.ipynb 8
@dataclass
class DomoAccount(DomoAccount_Default):
    pass

# %% ../../nbs/classes/50_DomoAccount.ipynb 10
class DomoAccount_Types(Enum):
    CREDENTIAL = DomoAccount_Credential
    DEFAULT = DomoAccount

    @classmethod
    def _missing_(cls, value):
        value = value.lower()

        for member in cls:
            if member.value == value:
                return member

            if value == "domo-access-token" and member.name == "CREDENTIAL":
                return member

        return cls.DEFAULT

    @classmethod
    def get(cls, value):

        if value == "domo-access-token":
            return cls.CREDENTIAL.value

        try:
            return cls[value].value

        except:
            return cls.DEFAULT.value

# %% ../../nbs/classes/50_DomoAccount.ipynb 13
@patch_to(DomoAccount, cls_method=True)
def _from_json(
    cls,
    obj: dict,
    is_admin_summary: bool = True,
    auth: dmda.DomoAuth = None,
    is_use_default_account_class=False,
    **kwargs
):
    """converts data_v1_accounts API response into an accounts class object"""

    if is_use_default_account_class:
        new_class = DomoAccount_Types.DEFAULT.value

    else:
        new_class = DomoAccount_Types.get(
            obj.get("dataProviderId") or obj.get("dataProviderType")
        )

    return new_class._class_from_json(
        obj=obj, is_admin_summary=is_admin_summary, auth=auth, **kwargs
    )

# %% ../../nbs/classes/50_DomoAccount.ipynb 25
class DomoAccounts_NoAccount(dmde.ClassError):
    def __init__(self, cls=None, cls_instance=None, message=None, domo_instance=None):
        super().__init__(
            cls=cls, cls_instance=cls_instance, message=message, entity_id=domo_instance
        )


@dataclass
class DomoAccounts:
    auth: dmda.DomoAuth = field(repr=False)

    accounts: List[DomoAccount] = None
    oauths : List[DomoAccount_OAuth] = None

    @staticmethod
    async def _get_accounts_accountsapi(
        auth: dmda.DomoAuth,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        is_use_default_account_class: bool = True,
        **kwargs,
    ):
        res = await account_routes.get_accounts(
            auth=auth, debug_api=debug_api, session=session
        )

        if return_raw:
            return res

        if len(res.response) == 0:
            return []

        return await dmce.gather_with_concurrency(
            n=60,
            *[
                DomoAccount.get_by_id(
                    account_id=json_obj.get("id"),
                    debug_api=debug_api,
                    session=session,
                    auth=auth,
                    is_use_default_account_class=is_use_default_account_class,
                    **kwargs,
                )
                for json_obj in res.response
            ],
        )

    @staticmethod
    async def _get_accounts_queryapi(
        auth: dmda.DomoAuth,
        debug_api: bool = False,
        additional_filters_ls=None,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        is_use_default_account_class: bool = False,
        **kwargs,
    ):
        """v2 api for works with group_account_v2 beta"""

        import domolibrary.routes.datacenter as datacenter_routes

        res = await datacenter_routes.search_datacenter(
            auth=auth,
            entity_type=datacenter_routes.Datacenter_Enum.ACCOUNT.value,
            additional_filters_ls=additional_filters_ls,
            session=session,
            debug_api=debug_api,
        )

        if return_raw:
            return res

        if len(res.response) == 0:
            return []

        return [
            DomoAccount._from_json(
                account_obj,
                auth=auth,
                is_use_default_account_class=is_use_default_account_class,
                **kwargs,
            )
            for account_obj in res.response
        ]

    async def get(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        is_use_default_account_class: bool = False,
        debug_num_stacks_to_drop: int = 2,
        **kwargs,
    ):
        domo_accounts = None
        try:
            domo_accounts = await self._get_accounts_queryapi(
                auth=self.auth,
                debug_api=debug_api,
                session=session,
                is_use_default_account_class=is_use_default_account_class,
                return_raw=return_raw,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
                **kwargs,
            )

        except datacenter_routes.SearchDatacenter_NoResultsFound as e:
            print(e)

        if not domo_accounts:
            domo_accounts = await self._get_accounts_accountsapi(
                auth=self.auth,
                debug_api=debug_api,
                session=session,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
                is_use_default_account_class=is_use_default_account_class,
                return_raw=return_raw,
                **kwargs,
            )

        if return_raw:
            return domo_accounts

        self.accounts = domo_accounts
        return self.accounts

# %% ../../nbs/classes/50_DomoAccount.ipynb 27
@patch_to(DomoAccounts)
async def get_oauths(
    self,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    debug_num_stacks_to_drop: int = 2,
):
    res = await account_routes.get_oauth_accounts(
        auth=self.auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__ .__name__,
    )

    if return_raw:
        return res
    
    return [ DomoAccount_OAuth._from_json(obj = obj, auth = self.auth) for obj in res.response]

    return res

# %% ../../nbs/classes/50_DomoAccount.ipynb 30
@patch_to(DomoAccounts, cls_method=True)
async def upsert_account(
    cls: DomoAccounts,
    auth: dmda.DomoAuth,

    account_id: str = None,
    account_name: str = None,
    account_config: AccountConfig = None,
    
    debug_api: bool = False,
    debug_prn: bool = False,
    return_raw: bool = False,
    return_search: bool = False,
    is_use_default_account_class: bool = True,
    session: httpx.AsyncClient = None,
    **kwargs,
):
    """search for an account and upsert it"""

    if not account_name and not account_id:
        raise UpsertAccount_MatchCriteria(domo_instance=auth.domo_instance)

    acc = None

    if account_id:
        try:
            acc = await DomoAccount.get_by_id(
                auth=auth,
                session=session,
                debug_api=debug_api,
                account_id=account_id,
                is_use_default_account_class=is_use_default_account_class,
                **kwargs,
            )
        except dmde.DomoError:
            pass

    if account_name and not acc:
        try:
            domo_accounts = DomoAccounts(auth =auth)
            await domo_accounts.get(
                debug_api=debug_api,
                session=session,
                is_use_default_account_class=is_use_default_account_class,
                **kwargs,
            )

            acc = next((da for da in domo_accounts.accounts if da.name.lower() == account_name.lower()), None)
        
        except dmde.DomoError:
            pass

    if return_search:
        return acc

    if not isinstance(acc, (DomoAccount_Default, DomoAccount, DomoAccount_Credential)):
        if debug_prn:
            print(f"creating {account_name} in {auth.domo_instance}")

        return await DomoAccount.create_account(
            account_name=account_name,
            config=account_config,
            auth=auth,
            debug_api=debug_api,
            return_raw=return_raw,
        )

    if account_name and account_id:
        if debug_prn:
            print(f"upsert-ing {acc.id} - {acc.name} in {auth.domo_instance}")

        await acc.update_name(
            account_name=account_name,
            debug_api=debug_api,
            return_raw=return_raw,
            session=session,
        )

    if account_config:  # upsert account
        acc.Config = account_config

        if debug_prn:
            print(f"upsertting {acc.id}:  updating config")

        await acc.update_config(
            debug_api=debug_api, return_raw=return_raw, session=session
        )

    return acc


@patch_to(DomoAccount_Default)
async def upsert_target_account(
    self,
    target_auth: dmda.DomoAuth,  # valid auth for target destination
    account_name: str = None,  # defaults to self.name
    debug_api: bool = False,
):
    """
    upsert an account in a target instance with self.Config
    """

    return await DomoAccounts.upsert_account(
        auth=target_auth,
        account_name=account_name or self.name,
        account_config=deepcopy(self.Config),
        debug_api=debug_api,
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAppDb.ipynb.

# %% auto 0
__all__ = ['to_json', 'AppDbDocument', 'AppDbCollection', 'AppDbCollections']

# %% ../../nbs/classes/50_DomoAppDb.ipynb 2
from typing import List, Callable
import httpx

import numbers
import asyncio

from dataclasses import field, dataclass
import domolibrary.client.DomoAuth as dmda
import datetime as dt
import domolibrary.utils.convert as dlcv
import domolibrary.utils.chunk_execution as ce

from nbdev.showdoc import patch_to

import domolibrary.routes.appdb as appdb_routes

# %% ../../nbs/classes/50_DomoAppDb.ipynb 6
def to_json(value):
    if hasattr(value, "to_json"):
        return value.to_json()

    if isinstance(value, dict):
        return {key: to_json(v) for key, v in value.items()}

    if isinstance(value, list):
        return [to_json(v) for v in value]

    if isinstance(value, numbers.Number):
        return value

    return str(value)


@dataclass
class AppDbDocument:

    auth: dmda.DomoAuth = field(repr=False)
    _collection_id: str

    _identity_columns: List[str]

    _id: str = None

    _created_on_dt: dt.datetime = None

    _updated_on_dt: dt.datetime = None

    content: dict = None

    def to_json(self, custom_content_to_json_fn: Callable = None):
        self.update_config()

        s = {"id": self._id, "collectionId": self._collection_id}

        if custom_content_to_json_fn:
            s.update({"content": custom_content_to_json_fn(self.content)})

        else:
            for key, value in self.__dict__.items():

                if key.startswith("_") or key in ["auth"]:
                    continue

                s.update({key: to_json(value)})

        return s

    def __eq__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False

        if self._identity_columns:
            return all(
                getattr(self, col) == getattr(other.col)
                for col in self._identity_columns
            )

        return self._id == other._id

    @classmethod
    def _from_json(
        cls,
        auth: dmda.DomoAuth,
        content,
        new_cls,
        identity_columns,
        collection_id=None,
        document_id=None,
        metadata=None,
        created_on_dt=None,
        updated_on_dt=None,
    ):

        if metadata:
            collection_id = metadata.pop("collectionId")

            created_on_dt = dlcv.convert_string_to_datetime(metadata.pop("createdOn"))

            updated_on_dt = dlcv.convert_string_to_datetime(metadata.pop("updatedOn"))
            document_id = metadata["id"]

        return new_cls(
            auth=auth,
            _id=document_id,
            _identity_columns=identity_columns,
            _collection_id=collection_id,
            _created_on_dt=created_on_dt,
            _updated_on_dt=updated_on_dt,
            content=content,
            **(content if cls.__name__ != "AppDbDocument" else {})
        )

    @classmethod
    def _from_api(
        cls,
        auth: dmda.DomoAuth,
        obj,
        identity_columns: List[str] = None,
    ):
        content = obj.pop("content")

        return cls._from_json(
            auth=auth,
            content=content,
            new_cls=cls,
            identity_columns=identity_columns,
            metadata=obj,
        )

    @classmethod
    def from_json(
        cls,
        auth: dmda.DomoAuth,
        collection_id: str,
        content: dict,
        identity_columns: List[str] = None,
    ):
        return cls._from_json(
            auth=auth,
            content=content,
            new_cls=cls,
            identity_columns=identity_columns,
            collection_id=collection_id,
        )

    def update_config(self):
        self.content = {
            key: value
            for key, value in self.__dict__.items()
            if key not in ["auth", "content"] and not key.startswith("_")
        }
        return self.content

    @classmethod
    async def get_by_id(
        cls,
        collection_id: str,
        document_id: str,
        auth: dmda.DomoAuth,
        identity_columns=None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=1,
        return_raw: bool = False,
    ):
        res = await appdb_routes.get_collection_document_by_id(
            auth=auth,
            collection_id=collection_id,
            document_id=document_id,
            parent_class=cls.__name__,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return cls._from_api(
            auth=auth,
            obj=res.response,
            identity_columns=identity_columns or [],
        )

# %% ../../nbs/classes/50_DomoAppDb.ipynb 9
@patch_to(AppDbDocument, cls_method = True)
async def create_document(
    cls,
    content : dict,
    collection_id : str,
    auth : dmda.DomoAuth,
    session: httpx.AsyncClient,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    return_raw : bool = False
):
    res = await appdb_routes.create_document(
        auth=auth,
        collection_id= collection_id,
        content = content,
        session=session,
        debug_api=debug_api,
        parent_class= cls.__name__,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    return await cls.get_by_id(collection_id = collection_id, document_id = res.response['id'], auth = auth)

# %% ../../nbs/classes/50_DomoAppDb.ipynb 10
@patch_to(AppDbDocument)
async def update_document(
    self,
    content : dict = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
    return_raw: bool = False
    
):
    res = await appdb_routes.update_document(
        auth=auth,
        collection_id=self._collection_id,
        document_id=self._id,
        content = content or self.to_json()['content'],
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=self.__class__.__name__,
    )

    if return_raw:    
        return res
        
    return res

# %% ../../nbs/classes/50_DomoAppDb.ipynb 15
@dataclass
class AppDbCollection:
    auth: dmda.DomoAuth = field(repr=False)
    id: str
    name: str

    created_on_dt: dt.datetime
    updated_on_dt: dt.datetime

    schema: dict

    domo_documents : List[AppDbDocument] = None

    @classmethod
    def _from_json(cls, auth, obj):

        return cls(
            auth=auth,
            id=obj["id"],
            name=obj["name"],
            created_on_dt=dlcv.convert_string_to_datetime(obj["createdOn"]),
            updated_on_dt=dlcv.convert_string_to_datetime(obj["updatedOn"]),
            schema=obj["schema"],
        )
        

    @classmethod
    async def get_by_id(
        cls,
        auth: dmda.DomoAuth,
        collection_id,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
    ):

        res = await appdb_routes.get_collection_by_id(
            auth=auth,
            collection_id=collection_id,
            parent_class=cls.__name__,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return cls._from_json(auth=auth, obj=res.response)


    def __eq__(self, other):
        if not isinstance(other, AppDbCollection):
            return False

        return self.id == other.id

# %% ../../nbs/classes/50_DomoAppDb.ipynb 16
@patch_to(AppDbCollection)
async def share_collection(
    self,
    domo_user=None,
    domo_group=None,
    permission: appdb_routes.Collection_Permission_Enum = appdb_routes.Collection_Permission_Enum.READ_CONTENT,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    session: httpx.AsyncClient = None,
):
    return await appdb_routes.modify_collection_permissions(
        collection_id=self.id,
        user_id=(domo_user and domo_user.id) or (await self.auth.who_am_i()).response["id"],
        group_id=domo_group and domo_group.id,
        permission=permission,
        auth=self.auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=self.__class__.__name__,
    )


@patch_to(AppDbCollection)
async def query_documents(
    self,
    query: dict = None,
    return_raw: bool = False,
    try_auto_share=False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    session: httpx.AsyncClient = None,
):

    documents = []
    loop_retry = 0
    while loop_retry <= 1 and not documents:
        try:
            res = await appdb_routes.get_documents_from_collection(
                auth=self.auth,
                collection_id=self.id,
                query=query,
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                session=session,
            )

            documents = res.response

        except appdb_routes.AppDb_GET_Exception as e:
            if try_auto_share:
                await self.share_collection(debug_api=debug_api)
                await asyncio.sleep(2)

            loop_retry += 1

            if loop_retry > 1:
                raise e

        if return_raw:
            return res

        self.domo_documents = await ce.gather_with_concurrency(
            *[
                AppDbDocument.get_by_id(
                    collection_id=self.id, document_id=doc["id"], auth=self.auth
                )
                for doc in documents
            ],
            n=60
        )

        return self.domo_documents

# %% ../../nbs/classes/50_DomoAppDb.ipynb 21
@patch_to(AppDbDocument, cls_method=True)
async def upsert(
    cls,
    auth: dmda.DomoAuth,
    collection_id,
    content: dict,
    identity_columns: List[str],
    session: httpx.AsyncClient = None,
    debug_api=False,
    debug_num_stacks_to_drop=3,
    return_raw: bool = False,
):
    domo_doc = None

    query = {f"content.{col}": content[col] for col in identity_columns}

    domo_collection = await AppDbCollection.get_by_id(
        auth=auth, collection_id=collection_id, return_raw=False
    )

    res = await domo_collection.query_documents(query=query, debug_api=debug_api)
    
    if res:
        domo_doc = res[0]

    if not domo_doc:
        return await cls.create_document(
            content=content,
            collection_id=collection_id,
            auth=auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=return_raw,
        )

    return await domo_doc.update_document(
        session=session, content=content, debug_api=debug_api, return_raw=return_raw
    )

# %% ../../nbs/classes/50_DomoAppDb.ipynb 25
@dataclass
class AppDbCollections:

    @classmethod
    async def get_collections(
        cls,
        auth: dmda.DomoAuth,
        datastore_id : str = None, 
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=1,
        return_raw: bool = False 
    ):

        res = await appdb_routes.get_collections(
            auth=auth,
            datastore_id= datastore_id,
            parent_class=cls.__class__.__name__,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw :
            return res
        
        return await ce.gather_with_concurrency(*[ AppDbCollection.get_by_id(collection_id = obj['id'], auth = auth) for obj in res.response], n= 10)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoJupyter.ipynb.

# %% ../../nbs/classes/50_DomoJupyter.ipynb 2
from __future__ import annotations

# %% auto 0
__all__ = ['DJW_Search_Error', 'DJW_InvalidClass', 'DomoJupyterWorkspace', 'DomoJupyterWorkspaces']

# %% ../../nbs/classes/50_DomoJupyter.ipynb 3
from domolibrary.routes.jupyter import (
    get_workspace_auth_token_params,
    JupyterAPI_Error,
    JupyterAPI_WorkspaceStarted,
)
from .DomoJupyter_Content import DomoJupyter_Content
from .DomoJupyter_DataSource import DomoJupyter_DataSource
from .DomoJupyter_Account import DomoJupyter_Account

# %% ../../nbs/classes/50_DomoJupyter.ipynb 4
import os
from dataclasses import dataclass, field
from typing import List, Any
import datetime as dt

import httpx

import domolibrary.utils.files as defi

import domolibrary.client.DomoAuth as dmda
import domolibrary.routes.jupyter as jupyter_routes
import domolibrary.classes.DomoAccount as dmac
import domolibrary.classes.DomoDataset as dmds
import domolibrary.classes.DomoUser as dmdu

import domolibrary.client.DomoError as dmde
import domolibrary.utils.chunk_execution as dmce

from nbdev.showdoc import patch_to

# %% ../../nbs/classes/50_DomoJupyter.ipynb 8
class DJW_Search_Error(dmde.ClassError):
    def __init__(self,cls = None, cls_instance = None, search_name = None, message=None, domo_instance = None):
        super().__init__(
            cls=cls,
            cls_instance=cls_instance,
            message=message or f"unable to find {search_name}",
            entity_id= domo_instance or (cls_instance and cls_instance.auth.domo_instance),
        )

class DJW_InvalidClass(dmde.ClassError):
    def __init__(self, cls_instance, message):
        super().__init__(cls_instance=cls_instance, message=message)


@dataclass
class DomoJupyterWorkspace:
    auth: dmda.DomoJupyterAuth = field(repr=False)
    id: str
    name: str
    description: str

    created_dt: dt.datetime
    updated_dt: dt.datetime

    owner: dict
    cpu: str
    memory: int

    last_run_dt: dt.datetime = None
    instances: List[dict] = None

    input_configuration: List[DomoJupyter_DataSource] = field(
        default_factory=lambda: []
    )
    output_configuration: List[DomoJupyter_DataSource] = field(
        default_factory=lambda: []
    )
    account_configuration: List[DomoJupyter_Account] = field(default_factory=lambda: [])
    content: List[DomoJupyter_Content] = field(default_factory=lambda: [])

    collection_configuration: List[dict] = None
    fileshare_configuration: List[dict] = None

    service_location: str = None
    service_prefix: str = None

    def __post_init__(self):
        self._update_auth_params()

        self.account_configuration.sort()
        self.output_configuration.sort()
        self.input_configuration.sort()

    def _update_auth_params(self):
        """extracts service location and prefix from "instance" object"""

        if self.instances:
            res = jupyter_routes.parse_instance_service_location_and_prefix(
                self.instances[0], self.auth.domo_instance
            )
            self.service_location = res["service_location"]
            self.service_prefix = res["service_prefix"]

    def update_auth_to_jupyter_auth(self, jupyter_token: str):
        """
        interacting with jupyter notebook content requires an additional authentication against the jupyter server (jupyter_token)
        currently we must manually scrape the token from the browser
        """

        self._update_auth_params()

        if isinstance(self.auth, dmda.DomoJupyterAuth):
            pass

        elif isinstance(self.auth, dmda.DomoFullAuth):
            self.auth = dmda.DomoJupyterFullAuth.convert_auth(
                auth=self.auth,
                service_location=self.service_location,
                jupyter_token=jupyter_token,
                service_prefix=self.service_prefix,
            )

        elif isinstance(self.auth, dmda.DomoTokenAuth):
            self.auth = dmda.DomoJupyterTokenAuth.convert_auth(
                auth=self.auth,
                service_location=self.service_location,
                jupyter_token=jupyter_token,
                service_prefix=self.service_prefix,
            )

        self.auth.generate_auth_header()

    @classmethod
    async def _from_json(
        cls,
        obj,
        auth,
        jupyter_token: str = None,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        is_use_default_account_class: bool = False,
        is_suppress_errors: bool = False,
    ):

        dj_workspace = cls(
            auth=auth,
            id=obj["id"],
            name=obj["name"],
            description=obj["description"],
            created_dt=obj["created"],
            updated_dt=obj["updated"],
            last_run_dt=obj.get("lastRun"),
            instances=obj["instances"],
            owner=obj["owner"],
            memory=obj["memory"],
            cpu=obj["cpu"],
            fileshare_configuration=obj["collectionConfiguration"],
        )

        if obj["outputConfiguration"]:
            dj_workspace.output_configuration = await dmce.gather_with_concurrency(
                *[
                    DomoJupyter_DataSource._from_json(obj=oc, dj_workspace=dj_workspace)
                    for oc in obj["outputConfiguration"]
                ],
                n=10
            )

        if obj["inputConfiguration"]:
            dj_workspace.input_configuration = await dmce.gather_with_concurrency(
                *[
                    DomoJupyter_DataSource._from_json(obj=ic, dj_workspace=dj_workspace)
                    for ic in obj["inputConfiguration"]
                ],
                n=10
            )

        if obj["accountConfiguration"]:
            dj_workspace.account_configuration = await dmce.gather_with_concurrency(
                *[
                    DomoJupyter_Account._from_json(
                        obj=ac,
                        dj_workspace=dj_workspace,
                        debug_api=debug_api,
                        session=session,
                        is_use_default_account_class=is_use_default_account_class,
                        is_suppress_errors=is_suppress_errors,
                    )
                    for ac in obj["accountConfiguration"]
                ],
                n=10
            )

        if jupyter_token:
            dj_workspace.update_auth_to_jupyter_auth(jupyter_token=jupyter_token)

        return dj_workspace

    def to_json(self):

        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "memory": int(self.memory),
            "cpu": self.cpu,
            "inputConfiguration": [
                confg.to_json() for confg in self.input_configuration or []
            ],
            "outputConfiguration": [
                confg.to_json() for confg in self.output_configuration or []
            ],
            "accountConfiguration": [
                confg.to_json() for confg in self.account_configuration or []
            ],
            "fileshareConfiguration": self.collection_configuration or [],
        }

    @classmethod
    async def get_by_id(
        cls,
        workspace_id,
        auth: dmda.DomoAuth,  # this API does not require the jupyter_token, but activities inside the workspace will require additional authentication
        jupyter_token=None,
        return_raw: bool = False,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        is_use_default_account_class: bool = False,
        is_suppress_errors: bool = False,
    ):

        res = await jupyter_routes.get_jupyter_workspace_by_id(
            workspace_id=workspace_id,
            auth=auth,
            session=session,
            debug_api=debug_api,
            parent_class=cls.__name__,
        )

        if return_raw:
            return res

        djw = await cls._from_json(
            auth=auth,
            obj=res.response,
            jupyter_token=jupyter_token,
            session=session,
            debug_api=debug_api,
            is_use_default_account_class=is_use_default_account_class,
            is_suppress_errors=is_suppress_errors,
        )

        return djw

# %% ../../nbs/classes/50_DomoJupyter.ipynb 10
@patch_to(DomoJupyterWorkspace, cls_method=True)
async def get_current_workspace(
    cls: DomoJupyterWorkspace,
    auth: dmda.DomoJupyterAuth,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    try:
        workspace_id = os.environ["DOMO_WORKSPACE_ID"]

    except KeyError as e:
        raise DJW_Search_Error(
            cls=cls,
            message=f"key error | workspace id {workspace_id} not found.  This only works in Domo's Jupyter Workspaces environment",
        ) from e

    return await cls.get_by_id(
        workspace_id=workspace_id, auth=auth, debug_api=debug_api, session=session
    )


@patch_to(DomoJupyterWorkspace)
async def get_account_configuration(
    self: DomoJupyterWorkspace,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    is_suppress_errors: bool = False,
    is_use_default_account_class: bool = False,
):
    djw = await self.get_by_id(
        workspace_id=self.id,
        auth=self.auth,
        session=session,
        debug_api=debug_api,
        is_suppress_errors=is_suppress_errors,
        is_use_default_account_class=is_use_default_account_class,
    )
    self.account_configuration = djw.account_configuration

    return self.account_configuration


@patch_to(DomoJupyterWorkspace)
async def get_input_configuration(
    self: DomoJupyterWorkspace,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    is_suppress_errors: bool = False,
):
    djw = await self.get_by_id(
        workspace_id=self.id,
        auth=self.auth,
        session=session,
        debug_api=debug_api,
        is_suppress_errors=is_suppress_errors,
    )
    self.input_configuration = djw.input_configuration

    return self.input_configuration


@patch_to(DomoJupyterWorkspace)
async def get_output_configuration(
    self: DomoJupyterWorkspace,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    is_suppress_errors: bool = False,
):
    djw = await self.get_by_id(
        workspace_id=self.id,
        auth=self.auth,
        session=session,
        debug_api=debug_api,
        is_suppress_errors=is_suppress_errors,
    )
    self.output_configuration = djw.output_configuration

    return self.output_configuration

# %% ../../nbs/classes/50_DomoJupyter.ipynb 12
@patch_to(DomoJupyterWorkspace)
def _add_config(self, config, attribute):

    config_ls = getattr(self, attribute)

    try:
        for idx, ex in enumerate(config_ls):
            if ex == config:
                config_ls[idx] = config

    except:
        config_ls.append(config)

    config_ls.sort()


@patch_to(DomoJupyterWorkspace)
def add_config_input_datasource(self, dja_datasource: DomoJupyter_DataSource):
    if not isinstance(dja_datasource, DomoJupyter_DataSource):
        raise DJW_InvalidClass(
            message="must passs instance of DomoJupyter_DataSource", cls_instance=self
        )

    return self._add_config(dja_datasource, attribute="input_configuration")


@patch_to(DomoJupyterWorkspace)
def add_config_output_datasource(self, dja_datasource: DomoJupyter_DataSource):

    if not isinstance(dja_datasource, DomoJupyter_DataSource):
        raise DJW_InvalidClass(
            message="must passs instance of DomoJupyter_DataSource", cls_instance=self
        )
    return self._add_config(dja_datasource, attribute="output_configuration")


@patch_to(DomoJupyterWorkspace)
def add_config_account(self, dja_account: DomoJupyter_Account):
    if not isinstance(dja_account, DomoJupyter_Account):
        raise DJW_InvalidClass(
            message="must passs instance of DomoJupyter_Account", cls_instance=self
        )
    return self._add_config(dja_account, attribute="account_configuration")

# %% ../../nbs/classes/50_DomoJupyter.ipynb 14
@dataclass
class DomoJupyterWorkspaces:
    auth: dmda.DomoAuth
    workspaces: List[DomoJupyterWorkspace] = None

    parent: Any = None

    @classmethod
    def _from_parent(cls, parent):
        return cls(parent=parent, auth=auth)

    async def get(
        self,
        is_use_default_account_class: bool = False,
        is_suppress_errors: bool = False,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
    ):
        res = await jupyter_routes.get_jupyter_workspaces(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.workspaces = await dmce.gather_with_concurrency(
            *[
                DomoJupyterWorkspace.get_by_id(
                    auth=self.auth,
                    workspace_id=workspace["id"],
                    debug_api=debug_api,
                    session=session,
                    is_use_default_account_class=is_use_default_account_class,
                    is_suppress_errors=is_suppress_errors,
                )
                for workspace in res.response
            ],
            n=10,
        )

        return self.workspaces

    async def search_workspace_by_name(
        self,
        workspace_name: str,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
        is_use_default_account_class: bool = False,
        is_suppress_errors: bool = False,
    ):
        res = await self.get(
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=True,
            is_suppress_errors=is_suppress_errors,
        )

        workspace = next(
            (
                workspace
                for workspace in res.response
                if workspace["name"].lower() == workspace_name.lower()
            ),
            None,
        )

        if not workspace:
            raise DJW_Search_Error(
                cls_instance=self, search_name=workspace_name
            )

        if return_raw:
            res.response = workspace
            return res

        return await DomoJupyterWorkspace.get_by_id(
            workspace_id=workspace["id"],
            auth=self.auth,
            session=session,
            debug_api=debug_api,
            is_use_default_account_class=is_use_default_account_class,
            is_suppress_errors=is_suppress_errors,
        )

# %% ../../nbs/classes/50_DomoJupyter.ipynb 18
@patch_to(DomoJupyterWorkspace)
async def update_config(
    self,
    config: dict = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=2,
):
    config = config or self.to_json()

    return await jupyter_routes.update_jupyter_workspace_config(
        auth=self.auth,
        workspace_id=self.id,
        config=config,
        parent_class=self.__class__.__name__,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )


@patch_to(DomoJupyterWorkspace)
async def add_account(
    self,
    domo_account: dmac.DomoAccount,
    dja_account: Any = None,
    domo_user: dmdu.DomoUser = None,
    is_update_config: bool = True,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):
    if not dja_account and not isinstance(domo_account, dmac.DomoAccount_Default):
        raise DJW_InvalidClass(
            message="must pass domo_account as class DomoAccount", cls_instance=self
        )

    dja_account = dja_account or DomoJupyter_Account(
        alias=domo_account.name, account_id=domo_account.id, dj_workspace=self
    )

    self.add_config_account(dja_account)

    if not is_update_config:
        return self.account_configuration

    share_user_id = (domo_user and domo_user.id) or (
        await self.auth.who_am_i()
    ).response["id"]

    await domo_account.Access.share(
        user_id=share_user_id,
        debug_api=debug_api,
        session=session,
    )

    retry = 0
    while retry <= 1:
        try:
            res = await self.update_config(debug_api=debug_api, session=session)

            if return_raw:
                return res

            account_obj = next(
                (
                    obj
                    for obj in res.response["accountConfiguration"]
                    if obj["alias"] == domo_account.name
                )
            )

            return DomoJupyter_Account(
                account_id=account_obj["account_id"],
                alias=account_obj["alias"],
                dj_workspace=self,
                domo_account=domo_account,
            )

        except JupyterAPI_Error as e:
            await domo_account.Access.share(
                user_id=share_user_id,
                debug_api=debug_api,
                session=session,
            )

            if retry == 1:
                raise e from e

            retry += 1


@patch_to(DomoJupyterWorkspace)
async def add_input_dataset(
    self,
    domo_dataset: dmds.DomoDataset,
    domojupyter_ds: _ = None,
    domo_user: dmdu.DomoUser = None,
    is_update_config: bool = True,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    if not domojupyter_ds and not isinstance(domo_dataset, dmds.DomoDataset):
        raise DJW_InvalidClass(
            message="must pass domo_dataset as class DomoDataset", cls_instance=self
        )

    domojupyter_ds = domojupyter_ds or DomoJupyter_DataSource(
        alias=domo_dataset.name, dataset_id=domo_dataset.id, dj_workspace=self
    )

    self.add_config_input_datasource(domojupyter_ds)

    if not is_update_config:
        return self.input_configuration

    domo_user = domo_user or await dmdu.DomoUser.get_by_id(
        auth=self.auth,
        user_id=(await self.auth.who_am_i()).response["id"],
        debug_api=debug_api,
        session=session,
    )

    retry = 0

    while retry <= 1:
        try:
            return await self.update_config(debug_api=debug_api, session=session)

        except JupyterAPI_Error as e:
            await domo_dataset.share(
                member=domo_user, auth=self.auth, debug_api=debug_api, session=session
            )

            if retry == 1 :
                raise e from e

            retry += 1


@patch_to(DomoJupyterWorkspace)
async def add_output_dataset(
    self,
    domo_dataset: dmds.DomoDataset,
    domojupyter_ds: DomoJupyter_DataSource = None,
    domo_user: dmdu.DomoUser = None,
    is_update_config: bool = True,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    if not domojupyter_ds and not isinstance(domo_dataset, dmds.DomoDataset):
        raise DJW_InvalidClass(
            message="must pass domo_dataset as class DomoDataset", cls_instance=self
        )

    domojupyter_ds = domojupyter_ds or DomoJupyter_DataSource(
        alias=domo_dataset.name, dataset_id=domo_dataset.id, dj_workspace=self
    )

    self.add_config_output_datasource(domojupyter_ds)

    if not is_update_config:
        return self.output_configuration

    domo_user = domo_user or await dmdu.DomoUser.get_by_id(
        auth=self.auth,
        user_id=(await self.auth.who_am_i()).response["id"],
        debug_api=debug_api,
        session=session,
    )

    retry = 0

    while retry <= 1:
        try:
            return await self.update_config(debug_api=debug_api, session=session)

        except JupyterAPI_Error as e:
            await domo_dataset.share(
                member=domo_user, auth=self.auth, debug_api=debug_api, session=session
            )

            if retry == 1:
                raise e from e

            retry += 1

# %% ../../nbs/classes/50_DomoJupyter.ipynb 19
@patch_to(DomoJupyterWorkspace)
async def get_content(
    self,
    debug_api: bool = False,
    return_raw: bool = False,
    is_recursive: bool = True,
    content_path: str = "",
):
    res = await jupyter_routes.get_content(
        auth=self.auth,
        debug_api=debug_api,
        content_path=content_path,
        debug_num_stacks_to_drop=2,
        parent_class=self.__class__.__name__,
        is_recursive=is_recursive,
        return_raw=return_raw,
    )

    if return_raw:
        return res

    self.content = [
        DomoJupyter_Content._from_json(obj, auth=self.auth)
        for obj in res.response
    ]
    return self.content

# %% ../../nbs/classes/50_DomoJupyter.ipynb 27
@patch_to(DomoJupyterWorkspace)
async def download_workspace_content(
    self, base_export_folder=None, replace_folder: bool = True
) -> str:
    """retrieves content from Domo Jupyter Workspace and downloads to a local folder"""

    base_export_folder = base_export_folder or f"{self.auth.domo_instance}/{self.name}"

    all_content = await self.get_content()
    all_content = [
        content for content in all_content if content.file_type != "directory"
    ]
    defi.upsert_folder(base_export_folder, replace_folder=replace_folder)

    return [
        content.export(default_export_folder=base_export_folder)
        for content in all_content
    ]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/instance_config_api_client.ipynb.

# %% auto 0
__all__ = ['ApiClient_GET_Error', 'get_api_clients', 'get_client_by_id', 'ApiClient_CRUD_Error', 'ApiClient_ScopeEnum',
           'create_api_client', 'ApiClient_RevokeError', 'revoke_api_client']

# %% ../../nbs/routes/instance_config_api_client.ipynb 2
from enum import Enum
from typing import List
import datetime as dt

import httpx

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde

# %% ../../nbs/routes/instance_config_api_client.ipynb 7
class ApiClient_GET_Error(dmde.RouteError):
    def __init__(self, res: rgd.ResponseGetData, message=None):

        super().__init__(res=res, message=message)


@gd.route_function
async def get_api_clients(
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class=None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/identity/v1/developer-tokens"

    res = await gd.get_data(
        url=url,
        method="GET",
        auth=auth,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )
    if not res.is_success:
        raise ApiClient_GET_Error(res=res)

    if return_raw:
        return res

    res.response = res.response["entries"]

    return res

# %% ../../nbs/routes/instance_config_api_client.ipynb 9
@gd.route_function
async def get_client_by_id(
    auth: dmda.DomoAuth,
    client_id: int,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class=None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:

    res = await get_api_clients(
        auth=auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        session=session,
        parent_class=parent_class,
    )

    if return_raw:
        return res

    client = next((obj for obj in res.response if obj.get("id") == int(client_id)))
    res.response = client
    return res


### NOT SUPPORTED

# @gd.route_function
# async def get_client_by_id(
#     auth: dmda.DomoAuth,
#     client_id : str,
#     debug_api: bool = False,
#     debug_num_stacks_to_drop=1,
#     parent_class=None,
#     session: httpx.AsyncClient = None,
#     return_raw: bool = False,
# ) -> rgd.ResponseGetData:

#     url = f"https://{auth.domo_instance}.domo.com/api/identity/v1/developer-tokens/{client_id}"

#     res = await gd.get_data(
#         url=url,
#         method="GET",
#         auth=auth,
#         debug_api=debug_api,
#         parent_class=parent_class,
#         num_stacks_to_drop=debug_num_stacks_to_drop,
#         session=session,
#     )
#     if not res.is_success:
#         raise ApiClient_GET_Error(res=res)

#     if return_raw:
#         return res

#     res.response = res.response["entries"]

#     return res

# %% ../../nbs/routes/instance_config_api_client.ipynb 12
class ApiClient_CRUD_Error(dmde.RouteError):
    def __init__(self, res: rgd.ResponseGetData, message: str = None):
        super().__init__(res=res, message=message)

class ApiClient_ScopeEnum(Enum):
    DATA = 'data'
    WORKFLOW = 'workflow'
    AUDIT = 'audit'
    BUZZ = 'buzz'
    USER = 'user'
    ACCOUNT = 'account'
    DASHBOARD = 'dashboard'

@gd.route_function
async def create_api_client(
    auth: dmda.DomoFullAuth, # usernme and password (full) auth required for this API
    client_name: str,
    client_description: str = f"generated via DL {str(dt.date.today()).replace('-', '')}",
    scope: List[ApiClient_ScopeEnum] = None,  # defaults to [data, audit]
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class=None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:
    
    if not isinstance(auth, dmda.DomoFullAuth):
        raise dmda.InvalidAuthTypeError(required_auth_type= dmda.DomoFullAuth)

    scope = scope or ["data", "audit"]
    
    if scope and isinstance(scope[0],ApiClient_ScopeEnum ):
        scope = [sc.value for sc in scope]
    


    url = f"https://api.domo.com/clients"

    headers = {"X-DOMO-CustomerDomain": f"{auth.domo_instance}.domo.com"}

    res = await gd.get_data(
        url=url,
        method="POST",
        auth=auth,
        body={"name": client_name, "description": client_description, "scope": scope},
        headers=headers,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )

    if not res.is_success:
        if res.status == 400:
            raise ApiClient_CRUD_Error(
                res, message=f"{res.response} -- does the client already exist?"
            )

        # if res.status == 403: # invalid auth type, but will be caught by the earlier test

        raise ApiClient_CRUD_Error(res)

    return res

# %% ../../nbs/routes/instance_config_api_client.ipynb 15
class ApiClient_RevokeError(dmde.RouteError):
    def __init__(
        self,
        res: rgd.ResponseGetData,
        client_id: str = None,
        message: str = None,
    ):
        super().__init__(
            res=res,
            message=message or f"error revoking client {client_id}",
        )


@gd.route_function
async def revoke_api_client(
    auth: dmda.DomoAuth,
    client_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class=None,
    session: httpx.AsyncClient = None,
):

    url = f"https://{auth.domo_instance}.domo.com/api/identity/v1/developer-tokens/{client_id}"

    res = await gd.get_data(
        url=url,
        method="DELETE",
        auth=auth,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )
    if not res.is_success:
        if res.status == 400:
            raise ApiClient_RevokeError(
                message=f"error revoking client {client_id}, validate that it exists.",
                res=res,
            )
        raise ApiClient_RevokeError(client_id=client_id, res=res)

    res.response = f"client {client_id} revoked"
    
    return res
